/*
 * Copyright 2013-2016 Sergey Ignatov, Alexander Zolotov, Florin Patan and 2016-2021 JetBrains s.r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.goide.completion;

import com.goide.GoCodeInsightFixtureTestCase;
import com.intellij.codeInsight.completion.CompletionType;
import com.intellij.codeInsight.lookup.LookupElement;
import com.intellij.codeInsight.lookup.LookupElementPresentation;
import com.intellij.testFramework.fixtures.TestLookupElementPresentation;
import com.intellij.util.containers.ContainerUtil;
import org.intellij.lang.annotations.MagicConstant;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.Objects;
import java.util.Set;

import static com.intellij.codeInsight.lookup.Lookup.*;

public abstract class GoCompletionTestBase extends GoCodeInsightFixtureTestCase {
  private GoCompletionTestFixture myCompletionFixture;

  @Override
  protected void setUp() throws Exception {
    super.setUp();
    myCompletionFixture = new GoCompletionTestFixture(this, myFixture);
  }

  @NotNull
  @Override
  protected String getBasePath() {
    return "completion";
  }

  protected void doTestCompletion() {
    myFixture.configureByFile(getTestName(true) + ".go");
    failOnFileLoading();
    myFixture.complete(getCompletionType());
    myFixture.checkResultByFile(getTestName(true) + "_after.go");
  }

  protected void doTestCompletion(@NotNull String itemToSelect) {
    doTestCompletion(getCompletionType(), itemToSelect);
  }

  protected void doTestCompletion(@NotNull CompletionType completionType, @NotNull String itemToSelect) {
    myFixture.configureByFile(getTestName(true) + ".go");
    failOnFileLoading();
    myFixture.complete(completionType);
    selectLookupItem(itemToSelect);
    myFixture.checkResultByFile(getTestName(true) + "_after.go");
  }

  protected void assertDeprecated(String... names) {
    failOnFileLoading();
    myFixture.complete(CompletionType.BASIC);
    LookupElement[] elements = myFixture.getLookupElements();
    assertNotNull(elements);
    Set<String> crossedOut = ContainerUtil.map2Set(ContainerUtil.filter(elements, e ->
      TestLookupElementPresentation.renderReal(e).isStrikeout()), LookupElement::getLookupString);
    assertSameElements(crossedOut, names);
  }

  protected void doTestVariants(@NotNull String txt, @NotNull CompletionType type, @NotNull CheckType checkType, String... variants) {
    myCompletionFixture.doTestVariants(txt, getDefaultFileName(), type, checkType, variants);
  }

  protected void doTestVariantsWithDefaultPresentation(@NotNull String txt, @NotNull CompletionType type, @NotNull CheckType checkType, String... variants) {
    myCompletionFixture.doTestVariantsWithDefaultPresentation(txt, getDefaultFileName(), type, checkType, variants);
  }

  public void doTestVariants(@NotNull CompletionType type, @NotNull GoCompletionTestBase.CheckType checkType, String... variants) {
    myCompletionFixture.doTestVariants(type, checkType, variants);
  }

  protected void doTestVariantsWithFile(@NotNull CheckType checkType, String... variants) {
    myCompletionFixture.doTestVariantsWithFile(getCompletionType(), checkType, variants);
  }

  protected void doTestVariantsWithFile(@NotNull CompletionType type, @NotNull CheckType checkType, String... variants) {
    myCompletionFixture.doTestVariantsWithFile(type, checkType, variants);
  }

  protected void doTestInclude(@NotNull String txt, String... variants) {
    myCompletionFixture.doTestVariants(txt, getDefaultFileName(), getCompletionType(), CheckType.INCLUDES, variants);
  }

  protected void doTestInclude(@NotNull String txt, int count, String... variants) {
    myCompletionFixture.doTestVariants(txt, getDefaultFileName(), getCompletionType(), count, CheckType.INCLUDES, variants);
  }

  protected void doTestExclude(@NotNull String txt, String... variants) {
    myCompletionFixture.doTestVariants(txt, getDefaultFileName(), getCompletionType(), CheckType.EXCLUDES, variants);
  }

  protected void doTestEquals(@NotNull String txt, String... variants) {
    myCompletionFixture.doTestVariants(txt, getDefaultFileName(), getCompletionType(), CheckType.EQUALS, variants);
  }

  @NotNull
  protected CompletionType getCompletionType() {
    return CompletionType.BASIC;
  }

  protected void doCheckResult(@NotNull String before, @NotNull String after) {
    myCompletionFixture.doCheckResult(before, after, null, getCompletionType(), getDefaultFileName());
  }

  protected void doCheckResult(@NotNull String before, @NotNull String after,
                               @Nullable @MagicConstant(flags = {
                                 AUTO_INSERT_SELECT_CHAR,
                                 COMPLETE_STATEMENT_SELECT_CHAR,
                                 NORMAL_SELECT_CHAR,
                                 REPLACE_SELECT_CHAR})
                               Character c) {
    myCompletionFixture.doCheckResult(before, after, c, getCompletionType(), getDefaultFileName());
  }

  protected void doCheckResult(@NotNull String before,
                               @NotNull String after,
                               @Nullable Character c,
                               @NotNull CompletionType completionType) {
    myCompletionFixture.doCheckResult(before, after, c, completionType, getDefaultFileName());
  }

  protected void doCheckResult(@NotNull String before, @NotNull String after, @NotNull String selectItem) {
    doCheckResult(before, after, selectItem, getCompletionType());
  }

  protected void doCheckResult(@NotNull String before,
                               @NotNull String after,
                               @NotNull String selectItem,
                               @NotNull CompletionType completionType) {
    myFixture.configureByText(getDefaultFileName(), before);
    failOnFileLoading();
    myFixture.complete(completionType);
    selectLookupItem(selectItem);
    myFixture.checkResult(after);
  }

  protected void selectLookupItem(@NotNull String selectItem) {
    LookupElement[] lookupElements = myFixture.getLookupElements();
    assertNotNull("Lookup is empty", lookupElements);
    LookupElement toSelect = null;
    for (LookupElement lookupElement : lookupElements) {
      if (selectItem.equals(lookupElement.getLookupString())) {
        toSelect = lookupElement;
        break;
      }
    }
    assertNotNull(selectItem + " not found in lookup", toSelect);
    myFixture.getLookup().setCurrentItem(toSelect);
    myFixture.type(NORMAL_SELECT_CHAR);
  }

  protected void selectLookupItem(@NotNull LookupElementPresentation presentation) {
    LookupElement toSelect = findLookupItem(presentation);
    assertNotNull(presentation + " not found in lookup", toSelect);
    myFixture.getLookup().setCurrentItem(toSelect);
    myFixture.type(NORMAL_SELECT_CHAR);
  }

  protected void selectLookupItem(int index) {
    LookupElement[] lookups = myFixture.getLookupElements();
    assertNotNull(lookups);
    assertTrue(index < lookups.length);
    myFixture.getLookup().setCurrentItem(lookups[index]);
    myFixture.type(NORMAL_SELECT_CHAR);
  }

  @Nullable
  protected LookupElement findLookupItem(@NotNull LookupElementPresentation presentation) {
    LookupElement[] lookupElements = myFixture.getLookupElements();
    assertNotNull("Lookup is empty", lookupElements);
    for (LookupElement lookupElement : lookupElements) {
      LookupElementPresentation current = TestLookupElementPresentation.renderReal(lookupElement);
      if (Objects.equals(presentation.getItemText(), current.getItemText()) &&
          Objects.equals(presentation.getTailText(), current.getTailText()) &&
          Objects.equals(presentation.getTypeText(), current.getTypeText())) {
        return lookupElement;
      }
    }
    return null;
  }

  @NotNull
  protected LookupElement getLookupItemByLookupString(@NotNull String lookupString) {
    LookupElement[] lookupElements = myFixture.getLookupElements();
    assertNotNull(lookupElements);
    LookupElement result = ContainerUtil.find(lookupElements, element -> lookupString.equals(element.getLookupString()));
    assertNotNull(result);
    return result;
  }

  protected static @NotNull LookupElementPresentation presentation(@NotNull String lookupString,
                                                                   @Nullable String tailText,
                                                                   @Nullable String typeText) {
    LookupElementPresentation p = new LookupElementPresentation();
    p.setItemText(lookupString);
    p.setTailText(tailText);
    p.setTypeText(typeText);
    return p;
  }

  protected String getDefaultFileName() {
    return "a.go";
  }

  public enum CheckType {EQUALS, INCLUDES, EXCLUDES, ORDERED_EQUALS}
}
