// Copyright 2016-2022 JetBrains s.r.o.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package com.goide.completion;

import com.goide.GoCodeInsightFixtureTestCase;
import com.goide.vgo.VgoUtil;
import com.intellij.codeInsight.completion.BaseCompletionLookupArranger;
import com.intellij.codeInsight.completion.CompletionType;
import com.intellij.codeInsight.lookup.LookupElement;
import com.intellij.codeInsight.lookup.LookupElementPresentation;
import com.intellij.openapi.util.text.StringUtil;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.testFramework.UsefulTestCase;
import com.intellij.testFramework.fixtures.CodeInsightTestFixture;
import com.intellij.util.Function;
import com.intellij.util.containers.ContainerUtil;
import org.intellij.lang.annotations.MagicConstant;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.stream.Collectors;

import static com.intellij.codeInsight.lookup.Lookup.*;
import static com.intellij.testFramework.UsefulTestCase.assertOrderedEquals;
import static com.intellij.testFramework.UsefulTestCase.assertSameElements;
import static junit.framework.TestCase.assertNotNull;
import static junit.framework.TestCase.assertTrue;

public class GoCompletionTestFixture {
  private final @NotNull UsefulTestCase myTestCase;
  private final @NotNull CodeInsightTestFixture myFixture;

  public GoCompletionTestFixture(@NotNull UsefulTestCase testCase, @NotNull CodeInsightTestFixture fixture) {
    myTestCase = testCase;
    myFixture = fixture;
  }

  @NotNull
  public CompletionType getCompletionType() {
    return CompletionType.BASIC;
  }

  @NotNull
  public String getDefaultFileName() {
    return "a.go";
  }

  public void doCheckResult(@NotNull String before, @NotNull String after) {
    doCheckResult(before, after, null);
  }

  public void doCheckResult(@NotNull String before, @NotNull String after,
                            @Nullable @MagicConstant(flags = {
                              AUTO_INSERT_SELECT_CHAR,
                              COMPLETE_STATEMENT_SELECT_CHAR,
                              NORMAL_SELECT_CHAR,
                              REPLACE_SELECT_CHAR})
                              Character c) {
    doCheckResult(before, after, c, getCompletionType(), getDefaultFileName());
  }

  public void doCheckResult(@NotNull String before,
                            @NotNull String after,
                            @Nullable Character c,
                            @NotNull CompletionType completionType,
                            @NotNull String defaultFileName) {
    myFixture.configureByText(defaultFileName, before);
    failOnFileLoading();
    myFixture.complete(completionType);
    if (c != null) myFixture.type(c);
    myFixture.checkResult(after);
  }

  public void doTestInclude(@NotNull String txt, String... variants) {
    doTestVariants(txt, getDefaultFileName(), getCompletionType(), GoCompletionTestBase.CheckType.INCLUDES, variants);
  }

  public void doTestVariants(@NotNull String txt,
                             @NotNull String defaultFileName,
                             @NotNull CompletionType type,
                             @NotNull GoCompletionTestBase.CheckType checkType,
                             String... variants) {
    doTestVariants(txt, defaultFileName, type, 1, checkType, GoCompletionTestFixture::getLookupString, variants);
  }

  public void doTestVariantsWithDefaultPresentation(@NotNull String txt,
                                                    @NotNull String defaultFileName,
                                                    @NotNull CompletionType type,
                                                    @NotNull GoCompletionTestBase.CheckType checkType,
                                                    String... variants) {
    doTestVariants(txt, defaultFileName, type, 1, checkType, GoCompletionTestFixture::getDefaultPresentationString, variants);
  }

  public void doTestVariants(@NotNull String txt,
                             @NotNull String defaultFileName,
                             @NotNull CompletionType type,
                             int count,
                             @NotNull GoCompletionTestBase.CheckType checkType,
                             String... variants) {
    myFixture.configureByText(defaultFileName, txt);
    failOnFileLoading(myFixture.findFileInTempDir(VgoUtil.GO_MOD));
    doTestVariantsInner(type, count, checkType, GoCompletionTestFixture::getLookupString, variants);
  }

  public void doTestVariants(@NotNull String txt,
                             @NotNull String defaultFileName,
                             @NotNull CompletionType type,
                             int count, 
                             @NotNull GoCompletionTestBase.CheckType checkType,
                             Function<LookupElement, String> lookupElementStringFunc,
                             String... variants) {
    myFixture.configureByText(defaultFileName, txt);
    failOnFileLoading(myFixture.findFileInTempDir(VgoUtil.GO_MOD));
    doTestVariantsInner(type, count, checkType, lookupElementStringFunc, variants);
  }

  public void doTestVariantsWithFile(@NotNull CompletionType type, @NotNull GoCompletionTestBase.CheckType checkType, String... variants) {
    myFixture.configureByFile(getTestName(true) + ".go");
    doTestVariants(type, checkType, variants);
  }

  public void doTestVariants(@NotNull CompletionType type, @NotNull GoCompletionTestBase.CheckType checkType, String... variants) {
    failOnFileLoading();
    doTestVariantsInner(type, 1, checkType, GoCompletionTestFixture::getLookupString, variants);
  }

  private void failOnFileLoading(@NotNull VirtualFile @NotNull... ignore) {
    GoCodeInsightFixtureTestCase.failOnFileLoading(myTestCase, myFixture, ignore);
  }

  @NotNull
  private String getTestName(boolean lowercaseFirstLetter) {
    return UsefulTestCase.getTestName(myTestCase.getName(), lowercaseFirstLetter);
  }

  private void doTestVariantsInner(@NotNull CompletionType type,
                                   int count,
                                   GoCompletionTestBase.CheckType checkType,
                                   Function<LookupElement, String> lookupString,
                                   String... variants) {
    myFixture.complete(type, count);
    LookupElement[] elements = myFixture.getLookupElements();
    List<String> stringList = elements == null
                              ? null
                              : ContainerUtil.map(elements, lookupString);

    assertNotNull("\nPossibly the single variant has been completed.\nFile after:\n" + myFixture
      .getFile().getText(), stringList);
    Collection<String> varList = new ArrayList<>(Arrays.asList(variants));
    if (checkType == GoCompletionTestBase.CheckType.ORDERED_EQUALS) {
      assertOrderedEquals(stringList, variants);
    }
    else if (checkType == GoCompletionTestBase.CheckType.EQUALS) {
      assertSameElements(stringList, variants);
    }
    else if (checkType == GoCompletionTestBase.CheckType.INCLUDES) {
      varList.removeAll(stringList);
      assertTrue("Missing variants: " + varList, varList.isEmpty());
    }
    else if (checkType == GoCompletionTestBase.CheckType.EXCLUDES) {
      varList.retainAll(stringList);
      assertTrue("Unexpected variants: " + varList, varList.isEmpty());
    }
  }

  private static String getLookupString(@NotNull LookupElement e) {
    return StringUtil.defaultIfEmpty(e.getLookupString(),
                                     e.getAllLookupStrings().stream().filter(s -> !s.isEmpty()).collect(Collectors.joining(", ")));
  }

  private static String getDefaultPresentationString(@NotNull LookupElement e) {
    LookupElementPresentation presentation = BaseCompletionLookupArranger.getDefaultPresentation(e);
    return presentation.getItemText() + presentation.getTailText();
  }
}