/*
 * Copyright 2013-2016 Sergey Ignatov, Alexander Zolotov, Florin Patan and 2016-2021 JetBrains s.r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.goide.quickfix;

import com.goide.GoCodeInsightFixtureTestCase;
import com.intellij.codeInsight.intention.IntentionAction;
import com.intellij.openapi.application.impl.NonBlockingReadActionImpl;
import com.intellij.openapi.fileEditor.FileDocumentManager;
import com.intellij.openapi.util.Condition;
import com.intellij.util.ArrayUtil;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.File;
import java.nio.file.Files;
import java.util.List;

import static com.intellij.openapi.util.io.FileUtil.toSystemDependentName;

public abstract class GoQuickFixTestBase extends GoCodeInsightFixtureTestCase {
  protected void doTest(@NotNull String quickFixName) {
    doTest(quickFixName, false, null);
  }

  protected void doTest(@NotNull String quickFixName, boolean checkHighlighting) {
    doTest(quickFixName, checkHighlighting, true);
  }

  protected void doTest(@NotNull String quickFixName, boolean checkHighlighting, boolean checkPreview) {
    doTest(quickFixName, checkHighlighting, checkPreview, null);
  }

  protected void doTestForExactFix(@NotNull String quickFixName) {
    doTestForExactFix(quickFixName, false);
  }

  protected void doTestForExactFix(@NotNull String quickFixName, boolean checkHighlighting) {
    doTestForExactFix(quickFixName, checkHighlighting, true);
  }

  protected void doTestForExactFix(@NotNull String quickFixName, boolean checkHighlighting, boolean checkPreview) {
    doTest(quickFixName, checkHighlighting, checkPreview, fix -> fix.getText().equals(quickFixName));
  }

  protected void doTest(@NotNull String testName,
                        @NotNull String quickFixName,
                        boolean checkHighlighting,
                        @Nullable Condition<? super IntentionAction> quickFixSelector) {
    doTest(testName, quickFixName, checkHighlighting, true, quickFixSelector);
  }

  protected void doTest(@NotNull String testName,
                        @NotNull String quickFixName,
                        boolean checkHighlighting,
                        boolean checkPreview,
                        @Nullable Condition<? super IntentionAction> quickFixSelector) {
    configure(checkHighlighting, testName);
    applySingleQuickFix(quickFixName, checkPreview, quickFixSelector);
    NonBlockingReadActionImpl.waitForAsyncTaskCompletion();
    myFixture.checkResultByFile(testName + "-after.go", true);
  }

  protected void doTest(@NotNull String quickFixName,
                        boolean checkHighlighting,
                        @Nullable Condition<? super IntentionAction> quickFixSelector) {
    doTest(quickFixName, checkHighlighting, true, quickFixSelector);
  }

  protected void doTest(@NotNull String quickFixName,
                        boolean checkHighlighting,
                        boolean checkPreview,
                        @Nullable Condition<? super IntentionAction> quickFixSelector) {
    doTest(getTestName(true), quickFixName, checkHighlighting, checkPreview, quickFixSelector);
  }

  protected void doTestNoFix(@NotNull String name) {
    doTestNoFix(name, false);
  }

  protected void doTestNoFix(@NotNull String name, boolean checkHighlighting) {
    configure(checkHighlighting, getTestName(true));
    List<IntentionAction> availableIntentions = myFixture.filterAvailableIntentions(name);
    assertEmpty(availableIntentions);
  }

  /**
   * Tests the output of applying the given quick fix on multiple files. This method is useful in cases when the main file (fileWithCaret)
   * imports elements from files in other packages, and we like to check the output of the multiple files. It is expected that the output of
   * the main file (fileWithCaret-after.go) be present. If any other file-after.go is present, the output of applying the quickfix for
   * that file will also be checked.
   */
  protected void doTestMultipleFiles(String quickfixName, String fileWithCaret, String...otherFiles) {
    String[] files = ArrayUtil.prepend(fileWithCaret, otherFiles);
    myFixture.configureByFiles(files);
    applySingleQuickFix(quickfixName, false);
    FileDocumentManager.getInstance().saveAllDocuments();
    myFixture.checkResultByFile(fileWithCaret, getFileAfterPath(fileWithCaret), true);
    for (String filePath : files) {
      String filePathAfterChange = getFileAfterPath(filePath);
      File fileAfter = new File(getTestDataPath(), toSystemDependentName(filePathAfterChange));
      if (Files.exists(fileAfter.toPath())) {
        myFixture.checkResultByFile(filePath, filePathAfterChange, true);
      }
    }
  }

  protected static String getFileAfterPath(String filePath) {
    int lastIndexOfSlash = filePath.lastIndexOf('/');
    return filePath.substring(0, lastIndexOfSlash) + filePath.substring(lastIndexOfSlash, filePath.lastIndexOf(".go")) + "-after.go";
  }

  protected void configure(boolean checkHighlighting, String testName) {
    if (checkHighlighting) {
      myFixture.testHighlighting(testName + ".go");
    }
    else {
      myFixture.configureByFile(testName + ".go");
      myFixture.doHighlighting();
    }
  }
}
