// Copyright 2016-2021 JetBrains s.r.o.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package com.goide.refactor.extractInterface;

import com.goide.GoCodeInsightFixtureTestCase;
import com.goide.psi.GoFile;
import com.goide.psi.GoNamedSignatureOwner;
import com.goide.psi.GoTypeSpec;
import com.goide.refactor.ui.GoDeclarationInfo;
import com.intellij.openapi.fileEditor.FileDocumentManager;
import com.intellij.openapi.util.text.StringUtil;
import com.intellij.psi.PsiFile;
import com.intellij.util.containers.CollectionFactory;
import one.util.streamex.StreamEx;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.List;
import java.util.Map;

public abstract class GoExtractInterfaceTestBase extends GoCodeInsightFixtureTestCase {

  protected void doExtractInterface(@NotNull String newTypeName, @NotNull PsiFile file, String... methodNames) {
    //noinspection ConstantConditions
    doExtractInterface(newTypeName, getDirectoryPath((GoFile)file), file.getName(), ((GoFile)file).getPackageName(), toMap(methodNames));
  }

  @SuppressWarnings("SameParameterValue")
  protected void doExtractInterface(@NotNull String newTypeName, @NotNull String fileName, @NotNull String packageName,
                                    String... methodNames) {
    String directoryPath = getDirectoryPath(getTypeSpecAtCaret().getContainingFile());
    doExtractInterface(newTypeName, directoryPath, fileName, packageName, toMap(methodNames));
  }

  @NotNull
  private static Map<String, String> toMap(String[] methodNames) {
    Map<String, String> map = CollectionFactory.createCaseInsensitiveStringMap();
    for (String methodName : methodNames) {
      map.put(methodName, methodName);
    }
    return map;
  }

  @NotNull
  private static String getDirectoryPath(@NotNull GoFile file) {
    //noinspection ConstantConditions
    return file.getContainingDirectory().getVirtualFile().getPath();
  }

  protected void doExtractInterface(@NotNull String newTypeName, @NotNull String directoryPath, @NotNull String fileName,
                                    @NotNull String packageName, Map<String, String> methodNames) {
    GoTypeSpec typeSpec = getTypeSpecAtCaret();

    String protocol = myFixture.getFile().getVirtualFile().getFileSystem().getProtocol();
    List<? extends GoNamedSignatureOwner> methods = GoExtractInterfaceDialog.getMethods(typeSpec);
    List<GoDeclarationInfo> infos = StreamEx.of(methods).map(GoDeclarationInfo::new).peek(info -> update(info, methodNames)).toList();

    GoExtractInterfaceDependencyGraph graph = new GoExtractInterfaceDependencyGraph(protocol, infos);
    GoExtractInterfaceInfoModel model = new GoExtractInterfaceInfoModel(graph);
    model.setTypeSpec(typeSpec);
    model.setRenameTypeSpec(false);
    model.setNewTypeName(newTypeName);
    model.setDirectoryPath(directoryPath);
    model.setFileName(StringUtil.trimEnd(fileName, ".go"));
    model.setPackageName(packageName);

    new GoExtractInterfaceProcessor(getProject(), model, false).run();
    FileDocumentManager.getInstance().saveAllDocuments();
  }

  private static void update(@NotNull GoDeclarationInfo info, @NotNull Map<String, String> methodNames) {
    String name = info.getDisplayName();
    info.setChecked(methodNames.isEmpty() || methodNames.containsKey(name));
    if (methodNames.containsKey(name)) {
      info.setExported(StringUtil.isCapitalized(methodNames.get(name)));
    }
  }

  @NotNull
  protected GoTypeSpec getTypeSpecAtCaret() {
    TypeSpecWithSelectedMethods spec = findSpecWithMethodAtOffset();
    assertNotNull("Type spec not found at caret", spec);
    return spec.myTypeSpec;
  }

  @Nullable
  protected TypeSpecWithSelectedMethods findSpecWithMethodAtOffset() {
    return GoExtractInterfaceHandler.findSpecWithMethodAtOffset(myFixture.getEditor(), myFixture.getFile());
  }
}
